import 'package:flutterx_utils/src/extensions/bool.dart';
import 'package:flutterx_utils/src/extensions/list.dart';
import 'package:flutterx_utils/src/utils/dto.dart';

extension MapExt<K, V> on Map<K, V> {
  /// Get enum entry from name at given [key] using [values] name-value mapping
  /// Set [lower] to true to be sure all keys of value are lowercase
  /// Also check [getEnumMap], [optEnumMap]
  /// E.g. json.optEnum('value_a', {'value_a': Values.valueA, 'value_b': Values.valueB})
  E requireEnumMap<E>(String key, Map<String, E> values, {bool lower = false}) {
    final entry = optEnumMap(key, values, lower: lower);
    if (entry != null) return entry;
    throw ArgumentError.value(this[key], key);
  }

  /// Get enum entry from name at given [key] using [values] name-value mapping
  /// Set [lower] to true to be sure all keys of value are lowercase
  /// Also check [optEnumMap], [requireEnumMap]
  /// E.g. json.optEnum('value_a', {'value_a': Values.valueA, 'value_b': Values.valueB})
  E getEnumMap<E>(String key, Map<String, E> values, {required E fallback, bool lower = false}) =>
      optEnumMap(key, values, lower: lower) ?? fallback;

  /// Get enum entry from name at given [key] using [values] name-value mapping
  /// Set [lower] to true to be sure all keys of value are lowercase
  /// Also check [getEnumMap], [requireEnumMap]
  /// E.g. json.optEnum('value_a', {'value_a': Values.valueA, 'value_b': Values.valueB})
  E? optEnumMap<E>(String key, Map<String, E> values, {bool lower = false}) {
    final value = this[key]?.toString();
    return value == null ? null : values[lower ? value.toLowerCase() : value];
  }

  /// Get enum entry from name at given [key] or throw [ArgumentError] if no entry matches name/index at given key.
  /// Set [ignoreCase] to true for non-case-sensitive name comparison
  /// Set [index] to true to read [key] as index instead of name
  /// Also check [getEnum], [optEnum]
  /// E.g. json.requireEnum('alignment', CrossAxisAlignment.values)
  E requireEnum<E>(String key, List<E> values, {bool ignoreCase = false, bool index = false}) {
    final entry = optEnum<E>(key, values, ignoreCase: ignoreCase, index: index);
    if (entry != null) return entry;
    throw ArgumentError.value(this[key], key);
  }

  /// Get enum entry from name at given [key] or [fallback] if no entry matches name/index at given key.
  /// Set [ignoreCase] to true for non-case-sensitive name comparison
  /// Set [index] to true to read [key] as index instead of name
  /// Also check [optEnum], [requireEnum]
  /// E.g. json.getEnum('alignment', CrossAxisAlignment.values, fallback: CrossAxisAlignment.start)
  E getEnum<E>(String key, List<E> values, {required E fallback, bool ignoreCase = false, bool index = false}) =>
      optEnum<E>(key, values, ignoreCase: ignoreCase, index: index) ?? fallback;

  /// Get enum entry from name at given [key] or null if no entry matches name/index at given key.
  /// Set [ignoreCase] to true for non-case-sensitive name comparison
  /// Set [index] to true to read [key] as index instead of name
  /// Also check [getEnum], [requireEnum]
  /// E.g. json.optEnum('alignment', CrossAxisAlignment.values)
  E? optEnum<E>(String key, List<E> values, {bool ignoreCase = false, bool index = false}) {
    if (index) {
      final i = this[key] as int;
      if (i < 0 || i >= values.length) return null;
      return values[i];
    }
    return values.optFromName(this[key]?.toString(), ignoreCase: ignoreCase);
  }

  /// Require int
  int requireInt(String key, {bool tryParse = false}) {
    final value = optInt(key, tryParse: tryParse);
    if (value != null) return value;
    throw ArgumentError.value(this[key], key);
  }

  /// Get int
  int getInt(String key, {int fallback = 0, bool tryParse = false}) => optInt(key, tryParse: tryParse) ?? fallback;

  /// Opt int
  int? optInt(String key, {bool tryParse = false}) {
    final value = this[key];
    return value == null
        ? null
        : value is num
            ? value.toInt()
            : tryParse
                ? num.tryParse(value.toString())?.toInt()
                : num.parse(value.toString()).toInt();
  }

  /// Require double
  double requireDouble(String key, {bool tryParse = false}) {
    final value = optDouble(key, tryParse: tryParse);
    if (value != null) return value;
    throw ArgumentError.value(this[key], key);
  }

  /// Get double
  double getDouble(String key, {double fallback = 0, bool tryParse = false}) =>
      optDouble(key, tryParse: tryParse) ?? fallback;

  /// Opt double
  double? optDouble(String key, {bool tryParse = false}) {
    final value = this[key];
    return value == null
        ? null
        : value is num
        ? value.toDouble()
        : tryParse
        ? double.tryParse(value.toString())
        : double.parse(value.toString());
  }

  /// Require bool
  bool requireBool(String key) {
    final value = optBool(key);
    if (value != null) return value;
    throw ArgumentError.value(this[key], key);
  }

  /// Get bool
  bool getBool(String key, {bool fallback = false}) => optBool(key) ?? fallback;

  /// Opt bool
  bool? optBool(String key) {
    final value = this[key];
    return value == null
        ? null
        : value is bool
        ? value
        : parseBool(value.toString());
  }

  /// Get entry value at given [key] or throw [ArgumentError] if value is not of type [E].
  /// E.g. json.require<int>('id') // throws if null
  E get<E>(String key) {
    final value = this[key];
    if (value is E) return value;
    throw ArgumentError.value(this[key], key);
  }

  E requireJsonObject<E>(String key, FromJsonObject<E> fromJson) {
    final value = optJsonObject<E>(key, fromJson);
    if (value != null) return value;
    throw ArgumentError.value(this[key], key);
  }

  E getJsonObject<E>(String key, FromJsonObject<E> fromJson, {required E fallback}) =>
      optJsonObject<E>(key, fromJson) ?? fallback;

  E? optJsonObject<E>(String key, FromJsonObject<E> fromJson) {
    final value = this[key];
    if (value is JsonObject) return fromJson(value);
    if (value == null) return null;
    throw StateError('Expected element of type $JsonObject at $key, found $value');
  }

  List<E> requireJsonList<E>(String key, FromJsonObject<E> fromJson) {
    final value = optJsonList<E>(key, fromJson);
    if (value != null) return value;
    throw ArgumentError.value(this[key], key);
  }

  List<E> getJsonList<E>(String key, FromJsonObject<E> fromJson, {List<E>? fallback}) =>
      optJsonList<E>(key, fromJson) ?? fallback ?? List<E>.empty();

  List<E>? optJsonList<E>(String key, FromJsonObject<E> fromJson) {
    final value = this[key];
    if (value is Iterable) return fromJsonList(value, fromJson);
    if (value == null) return null;
    throw StateError('Expected element of type $Iterable at $key, found $value');
  }

  /// Get parametrized list at given [key] from json map or [fallback] if specified or empty list of [E]
  /// throws a [StateError] if element at given key is not a list
  List<E> getList<E>(String key, {List<E>? fallback}) => optList<E>(key) ?? fallback ?? List<E>.empty();

  /// Get parametrized list at given [key] from json map or null if value is null
  /// throws a [StateError] if element at given key is not a list
  List<E>? optList<E>(String key) {
    final list = this[key];
    if (list == null) return null;
    if (list is List<E>) return list;
    if (list is List) return list.cast<E>();
    throw StateError('Expected element of type $List at $key, found ${list.runtimeType}');
  }

  /// Get parametrized list at given [key] from json map or throw [ArgumentError] if null
  /// throws a [StateError] if element at given key is not a list
  List<E> requireList<E>(String key) {
    final list = optList<E>(key);
    if (list != null) return list;
    throw ArgumentError.value(this[key], key);
  }

  /// Get all values where key satisfies [test] predicate
  Iterable<V> valuesWhereKey(bool Function(K key) test) sync* {
    for (final entry in entries) if (test(entry.key)) yield entry.value;
  }

  /// Get all keys where value satisfies [test] predicate
  Iterable<K> keysWhereValue(bool Function(V value) test) sync* {
    for (final entry in entries) if (test(entry.value)) yield entry.key;
  }
}
